module.exports = class Attribute {
    /**
     * Constructor
     * 
     * @param HTMLElement element
     * @param string attribute Passed only for custom attributes
     */
    constructor(tag, attribute = null) {
        this.tag = tag;
        this.attribute = attribute;
        this.init();
    }

    /**
     * Get then remove the attribute
     * 
     * @param   string attribute
     * @returns string
     */
    pull(attribute, defaultValue = null) {
        attribute = attribute.ltrim('[').rtrim(']');
        let attributeValue = this.get(attribute, defaultValue);

        this.remove(attribute);

        return attributeValue;
    }

    /**
     * Get value of the given attribute name
     * 
     * @param  string attribute
     * @returns string
     */
    get(attribute, $default = null) {
        attribute = attribute.ltrim('[').rtrim(']');
        return undefined !== this.tag.attributes[attribute] ? this.tag.attributes[attribute] : $default;
    }

    /**
     * Get the value from [] brackets
     * i.e [id]="this.mYid"
     * Usage: getProp('id')
     */
    getProp() {
        return this.get(...arguments);
    }

    /**
     * Get the attribute properly
     * if attribute="value" exists, return it,
     * otherwise, [attribute]="dynamicValue" will be returned
     */
    getOption() {
        return this.get(...arguments);
    }

    /**
     * Set the given attribute name and value
     */
    set(attribute, value) {
        this.tag.attributes[attribute] =  value;
    }

    /**
     * Get and remove attribute from list
     */
    forcePull() {
        return this.pull(...arguments);
    }

    /**
     * Get the attribute value and wrap it as a dynamic value
     * @param {*} attribute 
     */
    forcePullDynamic() {
        return this.pull(...arguments);
    }

    /**
     * Force pull the value of the given attribute/prop 
     */
    forcePullProp() {
        return this.pull(...arguments);
    }

    /**
     * Determine if the given attribute name exists
     * 
     * @param  string attribute
     * @returns bool
     */
    has(attribute) {
        return null !== this.get(attribute);
    }

    /**
     * Determine if any of the given attributes exists 
     * 
     * @param  array attributes
     * @returns bool
     */
    hasAny(...attributes) {
        for (let attribute of attributes) {
            if (this.has(attribute)) return true;
        }

        return false;
    }

    /**
     * Remove the given attribute name
     * 
     * @param  string attribute
     * @returns void
     */
    remove(attribute) {
        attribute = String(attribute).ltrim('[').rtrim(']');
        delete this.tag.attributes[attribute];
    }

    /**
     * Initialize the object
     */
    init() { }

    /**
     * Get attributes list
     */
    list() {
        return this.tag.attributes;
    }

    /**
     * Build final attributes and set it in the attributes list
     * 
     * @returns void
     */
    build() {
    }

    addClass(className, dynamicClass = false) {
        let classes = String(this.tag.attributes.class || '').trim('`');

        if (classes) {
            classes += ` `;
        }

        if (dynamicClass) {
            classes += '${' + className + '}';
        } else {
            classes += className;
        }

        this.tag.attributes.class = '`' + classes + '`';
    }
}