const Tag = require('./tag');
const TextNode = require('./text');
const { fs } = require('@flk/fs');
const { flatten } = require('./../helpers/flatten');
const { parseRoot, piper } = require('./../helpers/utils');

const { createElement, childElement } = require('./../helpers/builder');

const contentWrapper = require('./../pipelines/content-wrapper');
const thisToComponent = require('./../pipelines/this-to-component');
const controlStructure = require('./../pipelines/control-structure');
const navigationOperator = require('./../pipelines/navigator-operator');
const collectObservedPropertiesInComponent = require('./../pipelines/collect-observed-properties');

piper.pipeline(collectObservedPropertiesInComponent)
    .pipeline(thisToComponent)
    .pipeline(navigationOperator)
    .pipeline(contentWrapper)
    .pipeline(controlStructure);

// const collectedComponents >> core.js;

module.exports = class HtmlCompiler {
    /**
     * Constructor
     * 
     * @param  string html
     */
    constructor(html, viewName, parser = null) {
        this.objectName = 'comp' + random(3);
        this.objectName = 'component';

        this.viewName = viewName;

        this.observableProperties = [];

        this.parser = parser;

        this.originalCompiler = this;

        this.appendedProperties = null;

        this.rootParser = parseRoot(html, this);

        this.parsed = '';

        this.additionalArguments = [];

        this.children = {};

        this.componentsCallback = [];
    }

    /**
     * Get the original html file path
     * 
     * @return string 
     */
    getFilePath() {
        return this.originalCompiler.parser.filePath;
    }

    /**
     * Append the given properties to all children inside this html
     * This is usually used from child components
     * as it passes its state and key to all of its children
     * 
     * @param  {object} properties
     * @returns {void} 
     */
    appendToChildren(properties) {
        this.appendedProperties = properties;
    }

    /**
     * Observe the given property
     */
    observe(property) {
        if (Is.array(property)) {
            for (let prop of property) {
                this.observe(prop);
            }
            return;
        }

        let compiler = this.originalCompiler;
        if (compiler.observableProperties.includes(property)) return;

        compiler.observableProperties.push(property);
    }

    /**
     * Execute a callback each time a component is loaded
     * 
     * @param  callable callback
     */
    onComponentLoad(callback) {
        this.componentsCallback.push(callback);
    }

    /**
     * Get production code
     * Which mean basically with the try catch block and displaying some errors
     * 
     * @returns string
     */
    getCompiledCode() {
        if (this.observableProperties.includes('inputs')) {
            this.observableProperties = Array.remove(this.observableProperties, 'inputs');
        }

        let observableProperties = flatten(this.observableProperties);

        return `_comp({
            s: '${this.selector}', // s for selector
            c: '${this.component}', // c for component name
            u: ${this.isUnique}, // flag to determine if the component object will be unique or not
            o: ${observableProperties},
        }, function(${this.argumentsList()}) {
            ${this.parsed}
        })`;
    }

    /**
     * Get the arguments list that will be passed to the view handler
     *  
     */
    argumentsList() {
        let argumentsList = [this.objectName, ...this.additionalArguments];

        return argumentsList.join(', ');
    }

    /**
     * Start parsing
     */
    parse() {
        this.tags = this.rootParser.parse();

        const appendProperties = (children, properties) => {
            for (let child of children) {
                let newProperties = {};

                for (let property of HtmlCompiler.appendableProperties) {
                    if (property == 'key') {
                        let newKey = properties.key + (child.key || '');
                        newProperties[property] = child[property] = newKey;
                        continue;
                    }
                    if (properties[property]) {
                        newProperties[property] = child[property] = properties[property];
                    }
                }

                // if (properties.key) {
                //     newProperties.key = child.key = properties.key;
                // }

                // if (properties.state) {
                //     newProperties.state = child.state = properties.state;
                // }

                // if (properties.index) {
                //     newProperties.index = child.index = properties.index;
                // }

                // if (properties.insideLoop) {
                //     newProperties.insideLoop = child.insideLoop = properties.insideLoop;
                // }

                if (child.children) {
                    appendProperties(child.children, newProperties);
                }
            }
        };

        if (this.appendedProperties) {
            appendProperties(this.tags, this.appendedProperties);
        }

        if (! fs.isDirectory(ROOT + '/public/files')) {
            fs.makeDirectory(ROOT + '/public/files');
        } 
        fs.putJson(ROOT + '/public/files/' + this.viewName.replace(/\//g, '_') + '.json', this.tags);

        // if (this.viewName.includes('home-page')) {
        //     if (counter > 1) {
        //         die()
        //     }

        //     console.trace()

        //     counter++;
        // }

        this.parsed += this.extractChildren(this.tags);

        return this.parsed;
    }

    /**
     * Shorthand to create new element
     */
    createElement() {
        return createElement(this, ...arguments);
    }

    /**
     * Shorthand to create new child element
     */
    childElement() {
        return childElement(...arguments);
    }

    /**
     * Define component info
     */
    componentInfo(info) {
        this.component = info.component;
        this.isUnique = info.isUnique;
        this.selector = info.selector;
    }

    /**
     * Extract elements
     * 
     * @param   ParsedTag element
     * @returns string
     */
    extract(element, parentComponent = null, directParent = null) {
        if (element.type == 'comment') return '';

        let tagObject = null;

        if (element.type === 'text') {
            tagObject = new TextNode(element, this);
        } else {
            let tag = element.name;

            if (collectedComponents.has(tag)) {        
                let component = collectedComponents.get(tag),
                    ComponentHandler = component.handler;

                tagObject = new ComponentHandler(element, this);

                tagObject.parseContent = component.parseContent; // Parse the passed content to the component to html element
                tagObject.contentToString = component.contentToString; // pass the content as a property without parsing
                tagObject.component = component.component;
                tagObject.isUnique = component.isUnique;
                tagObject.isChild = component.isChild;
                tagObject.htmlFile = component.htmlFile;
            } else {
                tagObject = new Tag(element, this);
            }
        }
        
        if (parentComponent) {
            tagObject.parentComponent = parentComponent;
        }

        tagObject.directParent = directParent;

        return tagObject.build();
    }

    /**
     * Extract element children
     * 
     * @param   array
     * @returns void
     */
    extractChildren(children, parentComponent = null, directParent = null) {
        if (!children) return '';

        // if the passed element is an object not an array
        // and it holds the children key, then we'll iterate over it
        if (!Is.iterable(children) && children.children) {
            children = children.children;
        }

        let parsed = '';
        for (let child of children) {
            parsed += this.extract(child, parentComponent, directParent);
        }

        return parsed;
    }

    static get appendableProperties() {
        return [
            'key',
            'state',
            'index',
            'insideLoop',
        ];
    }
}