// don't forget to comment the DOMException in /node_modules/jsdom/lib/jsdom/living/helpers/validate-names.js:11:11

const Attribute = require('./attribute');

class Tag {
    /**
     * Constructor
     * 
     * @param ParsedElement element
     */
    constructor(element, htmlCompiler) {
        this.element = element;

        if (!element.children) {
            element.children = [];
        }

        this.htmlCompiler = htmlCompiler;
        this.originalAttributes = Object.assign({}, element.attributes);

        this.boot();
    }

    /**
     * Boot initial info
     */
    boot() {
        // when set to true, if any component passed an extra info such as events, attributes...etc to another component, 
        // the primary tag will get all these info
        this.isPrimary = false;
        this.style = {};
        this.userStyle = '';
        this.events = {};
        this.attributes = {};
        this.dynamicAttributesList = {}; // just an indicator to know if the attribute is dynamic or not
        this.onBuildEvents = [];
        this.onTerminateEvents = [];
        this.readyFunctionsList = []; // <element-name (ready)="doSomethingWhenElementIsRendered();"></element-name>

        this.attributes = {};

        this.specialAttributes = {};

        this.infoKey = 'this.props';

        this.setSpecialClassSyntax();

        for (let attribute in this.originalAttributes) {
            let value = this.originalAttributes[attribute];

            this.prepareAttribute(attribute, value);
        }

        this.collectCustomAttributes();

        this.init();
    }

    /**
     * Set the special syntax for the class [class] 
     */
    setSpecialClassSyntax() {
        let attrs = this.attrs();

        if (this.originalAttributes.class) {
            attrs.addClass(this.originalAttributes.class);

            delete this.originalAttributes.class;
        }

        if (this.originalAttributes['[class]']) {
            let conditionalClasses = this.originalAttributes['[class]'];

            delete this.originalAttributes['[class]'];

            attrs.addClass(`cls(${conditionalClasses})`, true);
        }
    }

    /**
     * Get the value for any special attribute
     * Special attributes are attributes start with *
     * i.e *for|*skip...etc
     * 
     * @param  {string} attribute 
     * @returns {string}
     */
    specialAttr(attribute) {
        return this.specialAttributes[attribute];
    }

    /**
     * Prepare the given attribute
     * 
     * @param  {string} attribute 
     * @param  {any} value 
     */
    prepareAttribute(attribute, value) {
        if (attribute.startsWith('[')) {
            this.setDynamicAttribute(attribute, value);
        } else if (attribute.startsWith('(')) {
            this.setEventsAttribute(attribute, value);
        } else if (attribute.startsWith('*')) {
            this.setSpecialAttribute(attribute, value);
        } else {
            this.setNormalAttribute(attribute, value);
        }
    }

    /**
     * Add normal attribute to attributes list
     * 
     * @param {string} attribute 
     * @param {string} value 
     */
    setNormalAttribute(attribute, value) {
        if (!Is.boolean(value) && !Is.numeric(value)) {
            value = '`' + value + '`';
        }

        this.attributes[attribute] = value;
    }

    /**
     * Special attributes like *as|*skip...etc
     * 
     * @param {string} attribute 
     * @param {string} value 
     */
    setSpecialAttribute(attribute, value) {
        this.specialAttributes[attribute.ltrim('*')] = value;
    }

    /**
     * Set the given attribute as event attribute
     * 
     * @param {string} attribute 
     * @param {string} value 
     */
    setEventsAttribute(attribute, value) {
        let eventAttribute = attribute.ltrim('(').rtrim(')');

        // ready event
        if (eventAttribute == 'ready') {
            return this.onElementReady(value);
        }

        // right click event
        eventAttribute = eventAttribute.replace(/rclick|rightclick|rightClick|right-click/, 'contextmenu');

        // multiple events in one attribute
        let events = eventAttribute.replace(/\(|\)/g, '').split(',').map(event => event.trim());

        // new: developer can now use multiple events with same callback
        // i.e (keyup,keydown)="console.log($el.value)"
        for (let event of events) {
            let fullEventName = 'on' + event.replace(/\(|\)/g, '').toCamelCase();
            this.addEventListener(fullEventName, value);
        }
    }

    /**
     * Set dynamic attribute
     * Dynamic attributes are attributes its name has []
     * i.e [class]="someVar"| [html]="this.someContent"
     *  
     * @param  {string} attribute 
     * @param  {string} value 
     */
    setDynamicAttribute(attribute, value) {
        attribute = attribute.ltrim('[').rtrim(']');
        this.attributes[attribute] = value;

        this.dynamicAttributesList[attribute] = value;
    }

    /**
     * Determine if the given attribute was passed as a dynamic one
     * i,e [label]="this.label"
     * 
     * @param   {string} attribute
     * @returns {boolean} 
     */
    isDynamicAttr(attribute) {
        return undefined !== this.dynamicAttributesList[attribute];
    }

    /**
     * Execute the given code on the current element is ready
     */
    onElementReady(callbackCode) {
        this.readyFunctionsList.push(callbackCode);

        return this;
    }

    /**
     * Get tag name
     * 
     * @returns string
     */
    tagName() {
        return this.element.name;
    }

    /**
     * An alias method to tagName
     */
    selector() {
        return this.tagName();
    }

    /**
     * Get attributes handler
     */
    attrs() {
        if (!this.attributesHandler) {
            this.attributesHandler = new Attribute(this);
        }

        return this.attributesHandler;
    }

    /**
     * Add new listener when the tag has been built
     * 
     * @param   callback callback
     * @returns void
     */
    onBuild(callback) {
        this.onBuildEvents.push(callback);
    }

    /**
     * Add new listener when the tag has been closed
     * 
     * @param   callback callback
     * @returns void
     */
    onTerminate(callback) {
        this.onTerminateEvents.push(callback);
    }

    /**
     * Add new event listener
     * @param string event 
     * @param string callback
     * @param boolean prepend >> if set to true, the event will be at the top of the callbacks
     */
    addEventListener(event, callback, prepend = false) {
        this.events[event] = Array.isArray(this.events[event]) ? this.events[event] : [];

        let method = prepend ? 'unshift' : 'push';

        this.events[event][method](callback);
    }

    /**
     * Add custom event listener
     * Usually set by custom attributes
     * 
     * @param string event
     * @param string action
     * @param boolean prepend >> if set to true, the event will be at the top of the callbacks
     */
    on(event, action, prepend = false) {
        if (event.includes(' ')) {
            let events = event.split(' ');
            for (let event of events) {
                this.on(event, action, prepend);
            }
            return;
        }

        event = 'on' + event;
        this.addEventListener(event, action, prepend);
    }

    /**
     * Initialize the object
     */
    init() { }

    /**
     * Call an attribute handler for the given attribute
     * 
     * @param  attribute
     * @returns Attribute
     */
    publishingAttribute(attribute) {
        let AttributeHandler = attributesList[attribute];

        if (!AttributeHandler) return;

        let attributeHandlerObject = new AttributeHandler(this, attribute);

        attributeHandlerObject.build();
    }

    /**
     * Collect custom attributes of the element
     */
    collectCustomAttributes() {
        for (let attribute in attributesList) {
            if (this.originalAttributes[attribute]) {
                this.publishingAttribute(attribute);
            }
        }

        for (let attributeList of bulkAttributesList) {
            let attributeHandlerObject = new attributeList(this),
                attributesList = attributeHandlerObject.map();

            for (let attribute in attributesList) {
                if (!attributeHandlerObject.has(attribute)) continue;

                let attributeHandler = attributesList[attribute];

                attributeHandler.call(attributeHandlerObject, attribute);
            }
        }
    }

    /**
     * Check if tag is a normal html tag or a component
     */
    type() {
        return 'tag';
    }

    /**
     * Get the name that will be passed to the callback function for 
     * rendering its element
     */
    asParentName() {
        return this.tagName().toCamelCase();
    }

    /**
     * Get inner html of the tag
     * 
     * @returns {string}
     */
    innerHTML() {
        return this.element.innerHTML;
    }

    /**
     * Build the tag
     */
    build() {
        let tagName = this.tagName();

        this.isPrimary = this.specialAttr('primary');

        if (this.specialAttr('skip')) {
            this.attributes.skip = true;
        }

        let children = '';

        if (this.attributes.html) {
            let content = this.attrs().pull('html');
            children = `elem(${content})`;
        } else {
            children = this.parseChildren();
        }

        let childrenList = '';

        if (children) {
            // if (!beforeAddingChildren && children.split('\n').length == 1 && children.startsWith('elem')) {
            //     childrenList = ',' + children.rtrim(';').ltrim('elem(').replace(/[\)]{1}$/, '');
            // } else {
            //     if (beforeAddingChildren) {
            //         beforeAddingChildren += '\n';
            //     }
            //     childrenList = `,(${this.asParentName()}) => {
            //         ${beforeAddingChildren}${children}
            //     }`;
            // }
        }

        let childrenOutput = this.childrenOutput(children);

        if (childrenOutput) {
            childrenList = `,${childrenOutput}`;
        }

        let attributes = `{`;

        let finalAttributes = this.getFinalAttributes();

        // finalAttributes.key = 'this.key';

        if (this.key()) {
            // finalAttributes.key += ' + ' + this.key();
            finalAttributes.key = this.key();
        }

        for (let attribute in finalAttributes) {
            let value = finalAttributes[attribute];

            attributes += `'${attribute}': ${value},`;
        }

        let styleList = '';

        if (!Is.empty(this.style)) {
            styleList = '{';

            for (let key in this.style) {
                let value = this.style[key];
                styleList += `'${key}': \`${value.trim('`')}\`,`;
            }

            styleList += '}';
        }

        if (!Is.empty(this.userStyle)) {
            if (styleList) {
                styleList = `Object.merge(${styleList}, ${this.userStyle})`;
            } else {
                styleList = this.userStyle;
            }
        }

        if (styleList) {
            attributes += `'style': ${styleList}`;
        }

        attributes += '}';

        if (attributes == '{}') {
            attributes = this.isPrimary ? this.infoKey : "''";
        } else {
            if (this.isPrimary) {
                attributes = `this.merge(${attributes})`;
            }
        }

        

        let elementFunction = `elem('${tagName}', ${attributes} ${childrenList})`;

        return this.output(elementFunction);
    }

    /**
     * Parse tag children
     * 
     * @returns {string}
     */
    parseChildren(parentComponent = this) {
        let output = '';

        for (let child of this.element.children) {
            if (child instanceof Tag) {
                output += child.build();
            } else {
                output += this.htmlCompiler.extract(child, parentComponent);
            }
        }

        return output;
    }

    /**
     * Get final attributes that will be passed to the element
     * 
     * @returns {object}
     */
    getFinalAttributes() {
        let attributes = this.attributes;

        for (let event in this.events) {
            let eventContent = `function (e) {
                let $el = this;
                `;
            for (let eventCallback of this.events[event]) {
                eventContent += eventCallback + '\n';
            }

            eventContent += '}';

            attributes[event] = eventContent;
        }

        for (let attribute in attributes) {
            if (!attribute.includes('-')) continue;

            let value = attributes[attribute];

            attributes[attribute.toCamelCase()] = value;
        }

        return attributes;
    }

    /**
     * Get the state for the tag
     * The state attribute appears only if the element is inside 
     * an if statement
     * 
     * @returns {boolean}
     */
    state() {
        return this.element.state;
    }

    /**
     * Get tag key
     */
    key() {
        return this.element.key ? '`' + this.element.key + '`' : null;
    }

    /**
     * Check if current tag is inside a for loop
     * 
     * @returns {boolean}
     */
    isInsideForLoop() {
        return this.element.insideLoop;
    }

    /**
     * Check if current tag is inside if statement
     * 
     * @returns {boolean}
     */
    isInsideIfStatement() {
        return Boolean(this.state());
    }

    /**
     * Get the final output of the tag
     * 
     * @param  {string} elementFunction
     * @returns {string} 
     */
    output(elementFunction) {
        let element = '',
            elementAlias = this.specialAttr('as');

        if (elementAlias) {
            if (this.element.insideLoop) {
                element += (`if (! ${this.htmlCompiler.objectName}.${elementAlias}) {
                        ${this.htmlCompiler.objectName}.${elementAlias} = [];
                    }

                    Array.pushOnce(${this.htmlCompiler.objectName}.${elementAlias}, ${elementFunction});`);
            } else {
                element += (`${this.htmlCompiler.objectName}.${elementAlias} = ${elementFunction};`);
            }
        } else {
            element = elementFunction + ';';
        }

        return element + '\n';
    }

    /**
     * Get the children output callback
     */
    childrenOutput(children) {
        let finalOutput = `${this.constructTag()}${children}${this.terminate()}`;

        if (!finalOutput) return '';

        return `(${this.asParentName()}) => {
            ${finalOutput}
        }`;
    }

    /**
     * This method is called just before adding the children
     * to the content callback
     */
    constructTag() {
        let output = '';

        for (let callback of this.onBuildEvents) {
            let returnedValue = Is.string(callback) ? callback : callback(this);

            if (Is.string(returnedValue)) {
                output += returnedValue;
            }
        }

        if (output) {
            output += '\n';
        }

        return output;
    }

    /**
     * Get tag variable name
     * Alias to asParentName
     */
    get variableName() {
        return this.asParentName();
    }

    /**
     * Add children to the tag
     * Children can be an element object or tag object
     * 
     * @param   {array} children
     * @returns {Tag}
     */
    addChildren(children) {
        this.element.children = this.element.children.concat(children);

        return this;
    }

    /**
     * Terminate the tag
     */
    terminate() {
        if (!Is.empty(this.readyFunctionsList)) {
            this.onTerminate(_ => (`
            // on element ready, this function will be called
            this.lr(_ => {
                let $el = ${this.asParentName()};
                ${this.readyFunctionsList.join('\n')}
            });                
`).trim()
            );
        }

        let output = '';
        for (let callback of this.onTerminateEvents) {
            let callbackOutput = Is.string(callback) ? callback : callback(this);

            if (Is.string(callbackOutput)) {
                output += callbackOutput + '\n';
            }
        }

        if (output) {
            output = '\n' + output;
        }

        return output;
    }

    /**
     * Get element object
     */
    getElement() {
        return this.element;
    }

    /**
     * Reset attributes except the given attributes list
     * 
     * @param  string...attributes
     */
    resetAttributes(...excludeAttributes) {
        for (let attribute in this.attributes) {
            if (excludeAttributes.includes(attribute)) continue;
            this.originalElement.setAttribute(attribute, this.attributes[attribute]);
        }
    }

    /**
     * Instead of rendering the current element
     * the current element will be inserted into other element and that other element 
     * will be rendered instead
     */
    injectInto(newParent, excludeAttributes = []) {
        let currentElement = this.originalElement,
            parent = currentElement.parentNode;
        // add original attributes once more
        this.resetAttributes(...excludeAttributes);

        // replace the child node
        // if (! parent) echo(currentElement.tagName)
        parent.insertBefore(newParent, currentElement);

        parent.removeChild(currentElement);

        newParent.appendChild(currentElement);

        // disable the build of the current tag as it will be built later
        this.build = () => { };

        this.htmlCompiler.extract(newParent);
    }
}

module.exports = Tag;